import os
import signal
import sys
import argparse
from PyQt5.QtWidgets import QApplication

from ortho_editor.drawing_app import DrawingApp


def editor():
    """Main function to initialize and run the drawing application."""
    parser = argparse.ArgumentParser(description="Drawing Tool with Smoothing")
    parser.add_argument(
        "--input_dir",
        default=os.getenv("ORTHO_RESULTS", "./results"),
        help="Directory containing input images (default: $ORTHO_RESULTS or ./results)",
    )
    args = parser.parse_args()

    app = QApplication(sys.argv)

    # Allow CTRL+C to interrupt the Qt event loop
    timer = app.startTimer(500)

    def timerEvent(event):
        app.processEvents()

    app.timerEvent = timerEvent

    window = DrawingApp(input_dir=args.input_dir)
    window.show()
    return app.exec_()

def signal_handler(signum, frame):
    """Handle CTRL+C by closing the Qt application"""
    from PyQt5.QtWidgets import QApplication

    QApplication.quit()

def main():
    # Set up signal handler for CTRL+C
    signal.signal(signal.SIGINT, signal_handler)
    sys.exit(editor())

if __name__ == "__main__":
    main()


from PyQt5.QtWidgets import (
    QToolBar,
    QAction,
    QPushButton,
    QSlider,
    QSpinBox,
    QButtonGroup,
    QToolButton,
)
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QColor, QFontDatabase, QFont
import os

# Material Icons constants
ICON_FOLDER_OPEN = "\ue2c8"
ICON_SAVE = "\ue161"
ICON_BRUSH = "\ue3ae"
ICON_LINE = "\ue922"
ICON_FORMAT_PAINT = "\ue243"
ICON_CIRCLE = "\ue3a6"
ICON_UNDO = "\ue166"
ICON_REDO = "\ue15a"
ICON_FILTER = "\ue3a5"
ICON_TUNE = "\ue15b"
ICON_AUTO_FIX = "\ue663"
ICON_ZOOM_IN = "\ue145"
ICON_ZOOM_OUT = "\ue15b"
ICON_ZOOM_FIT = "\ue56b"
ICON_ZOOM_100 = "\ue3b1"


class UIComponents:
    def __init__(self, main_window):
        super().__init__()
        self.main_window = main_window
        self.load_material_icons()
        self.setup_secondary_toolbar()
        self.setup_toolbar()
        if hasattr(self.main_window, "graphics_view"):
            self.main_window.graphics_view.update_pencil_cursor()

    def load_material_icons(self):
        """Load Material Icons font"""
        font_path = os.path.join(os.path.dirname(__file__), "fonts", "MaterialIcons-Regular.ttf")
        font_id = QFontDatabase.addApplicationFont(font_path)
        if font_id != -1:
            self.icon_font = QFont("Material Icons")
            self.icon_font.setPixelSize(24)
        else:
            print(f"Error: Could not load Material Icons font from {font_path}")
            self.icon_font = None

    def create_icon_action(self, icon_text, text, shortcut=None, tooltip=None):
        """Create an action with a material icon"""
        action = QAction(text, self.main_window)
        if self.icon_font:
            action.setFont(self.icon_font)
            action.setText(icon_text)
        if shortcut:
            action.setShortcut(shortcut)
        if tooltip:
            action.setToolTip(f"{tooltip} ({shortcut})" if shortcut else tooltip)
        return action

    def create_icon_button(self, icon_text, text, checkable=True):
        """Create a tool button with a material icon"""
        button = QToolButton()
        if self.icon_font:
            button.setFont(self.icon_font)
            button.setText(icon_text)
        button.setToolTip(text)
        if checkable:
            button.setCheckable(True)
        return button

    def setup_toolbar(self):
        """Create and setup the main toolbar with drawing tools"""
        self.toolbar = QToolBar()
        self.main_window.addToolBar(self.toolbar)

        # Create actions with icons
        self.open_action = self.create_icon_action(
            ICON_FOLDER_OPEN, "Open Directory", "Ctrl+O", "Open Directory"
        )
        self.save_action = self.create_icon_action(ICON_SAVE, "Save", "Ctrl+S", "Save")

        # Create tool buttons with icons
        self.pencil_button = self.create_icon_button(ICON_BRUSH, "Pencil")
        self.line_button = self.create_icon_button(ICON_LINE, "Line")
        self.fill_button = self.create_icon_button(ICON_FORMAT_PAINT, "Fill")
        self.eraser_button = self.create_icon_button(ICON_CIRCLE, "Eraser")

        # Create other actions with icons
        self.undo_action = self.create_icon_action(ICON_UNDO, "Undo", "Ctrl+Z", "Undo")
        self.redo_action = self.create_icon_action(
            ICON_REDO, "Redo", "Ctrl+Shift+Z", "Redo"
        )
        self.smooth_action = self.create_icon_action(
            ICON_FILTER, "Smooth", None, "Smooth"
        )
        self.thin_action = self.create_icon_action(ICON_TUNE, "Thin", None, "Thin")
        self.smooth_thin_smooth_action = self.create_icon_action(
            ICON_AUTO_FIX, "Smooth-Thin-Smooth", None, "Smooth-Thin-Smooth"
        )

        # Add actions and buttons to toolbar
        self.toolbar.addAction(self.open_action)
        self.toolbar.addAction(self.save_action)
        self.toolbar.addSeparator()
        self.toolbar.addWidget(self.pencil_button)
        self.toolbar.addWidget(self.line_button)
        self.toolbar.addWidget(self.fill_button)
        self.toolbar.addSeparator()
        self.toolbar.addWidget(self.eraser_button)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.undo_action)
        self.toolbar.addAction(self.redo_action)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.smooth_action)
        self.toolbar.addAction(self.thin_action)
        self.toolbar.addAction(self.smooth_thin_smooth_action)
        self.toolbar.addSeparator()

        # Add size controls
        self.size_label = QAction("Size:", self.main_window)
        self.toolbar.addAction(self.size_label)

        self.setup_size_controls()
        self.setup_size_presets()

        self.toolbar.addSeparator()

        self.setup_color_button()
        self.update_color_button(QColor(255, 165, 0))

    def setup_secondary_toolbar(self):
        """Create and setup a secondary toolbar for opacity and zoom controls"""
        self.secondary_toolbar = QToolBar()
        # Add the secondary toolbar to the bottom of the window
        self.main_window.addToolBar(Qt.BottomToolBarArea, self.secondary_toolbar)

        # Add opacity controls
        self.opacity_label = QAction("Opacity:", self.main_window)
        self.secondary_toolbar.addAction(self.opacity_label)

        self.opacity_slider = QSlider(Qt.Horizontal)
        self.opacity_slider.setMinimum(0)
        self.opacity_slider.setMaximum(100)
        self.opacity_slider.setValue(50)
        self.opacity_slider.setFixedWidth(100)
        self.secondary_toolbar.addWidget(self.opacity_slider)

        self.secondary_toolbar.addSeparator()

        # Add zoom controls with proper actions
        self.zoom_label = QAction("Zoom:", self.main_window)
        self.secondary_toolbar.addAction(self.zoom_label)

        self.zoom_out_action = self.create_icon_action(
            ICON_ZOOM_OUT, "Zoom Out", "Ctrl+-", "Zoom Out"
        )
        self.secondary_toolbar.addAction(self.zoom_out_action)

        self.zoom_level_label = QAction("100%", self.main_window)
        self.zoom_level_label.setEnabled(False)
        self.secondary_toolbar.addAction(self.zoom_level_label)

        self.zoom_in_action = self.create_icon_action(
            ICON_ZOOM_IN, "Zoom In", "Ctrl++", "Zoom In"
        )
        self.secondary_toolbar.addAction(self.zoom_in_action)

        self.zoom_fit_action = self.create_icon_action(
            ICON_ZOOM_FIT, "Fit", None, "Fit to Window"
        )
        self.secondary_toolbar.addAction(self.zoom_fit_action)

        self.zoom_100_action = self.create_icon_action(
            ICON_ZOOM_100, "100%", "Ctrl+0", "Actual Size"
        )
        self.secondary_toolbar.addAction(self.zoom_100_action)

    def setup_size_controls(self):
        """Setup size spinbox and slider"""
        self.size_spinbox = QSpinBox()
        self.size_spinbox.setMinimum(4)
        self.size_spinbox.setMaximum(100)
        self.size_spinbox.setValue(8)
        self.size_spinbox.setFixedWidth(60)
        self.toolbar.addWidget(self.size_spinbox)

        self.size_slider = QSlider(Qt.Horizontal)
        self.size_slider.setMinimum(4)
        self.size_slider.setMaximum(100)
        self.size_slider.setValue(8)
        self.size_slider.setFixedWidth(100)
        self.toolbar.addWidget(self.size_slider)

        self.toolbar.addSeparator()

    def setup_size_presets(self):
        """Setup preset size buttons"""
        self.size_button_group = QButtonGroup()
        self.size_button_group.setExclusive(True)

        # Define preset sizes
        presets = [4, 8, 16, 32]

        for size in presets:
            btn = QPushButton(str(size))
            btn.setCheckable(True)
            btn.setFixedSize(32, 32)
            if size == 8:  # Default size
                btn.setChecked(True)
            self.toolbar.addWidget(btn)
            self.size_button_group.addButton(btn)
            btn.clicked.connect(lambda checked, s=size: self.update_size_controls(s))

    def update_size_controls(self, size):
        """Update all size controls to match the given size"""
        self.size_slider.setValue(size)
        self.size_spinbox.setValue(size)

        # Update button states
        for button in self.size_button_group.buttons():
            button.setChecked(int(button.text()) == size)

    def setup_color_button(self):
        """Setup color picker button"""
        self.color_button = QPushButton()
        self.color_button.setFixedSize(32, 32)
        self.toolbar.addWidget(self.color_button)
        # Color will be set when tool is selected

    def update_color_button(self, color):
        """Update color button appearance"""
        self.color_button.setStyleSheet(
            f"background-color: {color.name()}; border: 2px solid #666666;"
        )

    def update_tool_specific_controls(self, tool):
        """Update controls based on tool settings"""
        settings = self.main_window.state.current_tool_settings
        self.update_size_controls(settings["size"])
        self.update_color_button(settings["color"])
        self.color_button.setEnabled(tool != "Eraser")



import os
import numpy as np
from PyQt5.QtWidgets import (
    QMainWindow,
    QVBoxLayout,
    QWidget,
    QGraphicsScene,
    QHBoxLayout,
    QListWidget,
    QFileDialog,
    QColorDialog,
)
from PyQt5.QtGui import QPixmap, QImage, QColor
from PyQt5.QtCore import Qt

from .custom_graphics_view import CustomGraphicsView
from .undo_manager import UndoManager
from .image_utils import ImageUtils
from .ui_components import UIComponents
from .image_processor import ImageProcessor
from .app_state import AppState


class DrawingApp(QMainWindow):
    def __init__(self, input_dir="./results"):
        super().__init__()
        # Initialize state manager
        self.state = AppState()
        # Connect state signals
        self.connect_state_signals()
        self.init_window()
        self.init_attributes(input_dir)
        self.init_ui_components()
        self.init_directories()
        self.load_images_list()

    # ---- Initialization Methods ----

    def init_window(self):
        """Initialize window properties"""
        self.setWindowTitle("Drawing Tool with Smoothing")
        self.resize(1280, 800)

        # Initialize central widget with horizontal layout
        self.central_widget = QWidget()
        self.setCentralWidget(self.central_widget)
        self.layout = QHBoxLayout(self.central_widget)

    def init_attributes(self, input_dir):
        """Initialize class attributes"""
        self.input_dir = input_dir
        self.current_tool = "Line"
        self.image = None
        self.background_image = None
        self.pixmap_item = None
        self.background_item = None
        self.current_filename = "mask-out.png"
        self.has_unsaved_changes = False
        self.undo_manager = UndoManager()

    def init_ui_components(self):
        """Initialize UI components and layouts"""
        # Create sidebar
        self.setup_sidebar()

        # Create drawing area
        self.setup_drawing_area()

        # Initialize UI components and connect actions
        self.ui = UIComponents(self)
        self.connect_actions()

    def setup_sidebar(self):
        """Setup sidebar for image list"""
        self.sidebar = QListWidget()
        self.sidebar.setMaximumWidth(200)
        self.sidebar.itemClicked.connect(self.load_background_from_list)
        self.layout.addWidget(self.sidebar)

    def setup_drawing_area(self):
        """Setup drawing container and graphics view"""
        self.drawing_container = QWidget()
        self.drawing_layout = QVBoxLayout(self.drawing_container)
        self.layout.addWidget(self.drawing_container)

        self.graphics_view = CustomGraphicsView(self)
        self.scene = QGraphicsScene()
        self.graphics_view.setScene(self.scene)
        self.drawing_layout.addWidget(self.graphics_view)

    def init_directories(self):
        """Initialize required directories"""
        directories = [
            self.input_dir,
            os.path.join(self.input_dir, "inputs"),
            os.path.join(self.input_dir, "targets"),
        ]
        for directory in directories:
            os.makedirs(directory, exist_ok=True)

    # ---- Action Connections ----

    def connect_actions(self):
        """Connect all UI actions to their handlers"""
        # File actions
        self.ui.open_action.triggered.connect(self.open_directory)
        self.ui.save_action.triggered.connect(self.save_image)

        # Tool buttons
        self.ui.pencil_button.clicked.connect(lambda: self.select_tool("Pencil"))
        self.ui.eraser_button.clicked.connect(lambda: self.select_tool("Eraser"))
        self.ui.line_button.clicked.connect(lambda: self.select_tool("Line"))
        self.ui.fill_button.clicked.connect(lambda: self.select_tool("Fill"))

        # Edit actions
        self.ui.undo_action.triggered.connect(self.undo)
        self.ui.redo_action.triggered.connect(self.redo)
        self.ui.smooth_action.triggered.connect(self.smooth_image)
        self.ui.thin_action.triggered.connect(self.thin_image)
        self.ui.smooth_thin_smooth_action.triggered.connect(
            self.smooth_thin_smooth_image
        )

        # Style actions
        self.ui.color_button.clicked.connect(self.choose_color)

        # Size control connections
        self.connect_size_controls()

        # Opacity control connection
        self.ui.opacity_slider.valueChanged.connect(self.change_opacity)

        # Connect zoom controls
        self.ui.zoom_in_action.triggered.connect(self.graphics_view.zoom_in)
        self.ui.zoom_out_action.triggered.connect(self.graphics_view.zoom_out)
        self.ui.zoom_fit_action.triggered.connect(self.graphics_view.zoom_fit)
        self.ui.zoom_100_action.triggered.connect(self.graphics_view.zoom_100)

        # Add keyboard shortcuts for zoom
        self.ui.zoom_in_action.setShortcut("Ctrl++")
        self.ui.zoom_out_action.setShortcut("Ctrl+-")
        self.ui.zoom_100_action.setShortcut("Ctrl+0")

    def connect_size_controls(self):
        """Connect size-related controls"""
        self.ui.size_slider.valueChanged.connect(self.ui.size_spinbox.setValue)
        self.ui.size_spinbox.valueChanged.connect(self.ui.size_slider.setValue)
        self.ui.size_slider.valueChanged.connect(self.change_pen_size)
        self.ui.size_spinbox.valueChanged.connect(self.change_pen_size)
        self.ui.size_slider.valueChanged.connect(self.ui.update_size_controls)
        self.ui.size_spinbox.valueChanged.connect(self.ui.update_size_controls)

    def connect_state_signals(self):
        """Connect state change signals to UI updates"""
        self.state.image_changed.connect(self.update_scene)
        self.state.unsaved_changes.connect(self.update_window_title)

    # ---- File Operations ----

    def load_images_list(self):
        """Load image files from input directory into sidebar"""
        self.sidebar.clear()
        inputs_dir = os.path.join(self.input_dir, "inputs")

        if os.path.exists(inputs_dir):
            for filename in os.listdir(inputs_dir):
                if filename.lower().endswith((".png", ".jpg", ".jpeg", ".bmp")):
                    input_path = os.path.join(inputs_dir, filename)
                    if os.path.exists(input_path):
                        self.sidebar.addItem(filename)

    def get_file_path(self, title="Open Image", directory=""):
        """Helper method to get file path from dialog"""
        return QFileDialog.getOpenFileName(
            self, title, directory, "Images (*.png *.jpg *.bmp)"
        )[0]

    def load_image_from_path(self, file_path):
        """Helper method to load and process image from path"""
        if not file_path or not os.path.exists(file_path):
            return

        # Load image and convert to numpy array
        temp_image = QImage(file_path)
        array = ImageUtils.qimage_to_numpy(temp_image)

        # Convert to grayscale
        grayscale = np.mean(array[:, :, :3], axis=2)

        # Threshold to black and white (binary)
        binary = (grayscale > 128).astype(np.uint8)

        # Create RGBA array
        height, width = binary.shape
        rgba = np.zeros((height, width, 4), dtype=np.uint8)

        # Set yellow color (255, 255, 0) where binary is 1
        rgba[binary == 1] = [255, 255, 0, 255]

        # Convert back to QImage
        self.image = QImage(
            rgba.data, width, height, width * 4, QImage.Format_RGBA8888
        ).copy()

        if self.background_image and (
            self.image.size() != self.background_image.size()
        ):
            # Resize drawing layer to match background
            self.image = self.image.scaled(self.background_image.size())

    def load_background_from_path(self, file_path):
        """Helper method to load background from path"""
        if not file_path or not os.path.exists(file_path):
            return

        self.background_image = QImage(file_path).convertToFormat(QImage.Format_RGB32)
        # If no drawing layer exists, create a transparent one
        if self.image is None:
            self.image = QImage(self.background_image.size(), QImage.Format_ARGB32)
            self.image.fill(Qt.transparent)
        self.update_scene()

    def save_image(self):
        if self.image:
            save_filename = (
                self.current_filename
                if hasattr(self, "current_filename")
                else "mask-out.png"
            )
            file_path = os.path.join(self.input_dir, "targets", save_filename)

            # Convert to binary image and save
            binary_image = ImageProcessor.create_binary_save_image(self.image)
            if binary_image:
                binary_image.save(file_path)
                self.has_unsaved_changes = False
                self.update_window_title()

                # Update the image view with yellow pixels
                array = ImageUtils.qimage_to_numpy(self.image)
                height, width = array.shape[:2]
                rgba = np.zeros((height, width, 4), dtype=np.uint8)

                # Set yellow color (255, 255, 0) where alpha channel is not 0
                mask = array[:, :, 3] > 0
                rgba[mask] = [255, 255, 0, 255]

                # Convert back to QImage and update the view
                self.image = QImage(
                    rgba.data, width, height, width * 4, QImage.Format_RGBA8888
                ).copy()
                self.update_scene()

    def load_background_from_list(self, item):
        """Load background image and corresponding drawing when clicked in sidebar"""
        if not item:
            return

        # Clear undo/redo history before loading new image
        self.undo_manager = UndoManager()

        self.current_filename = item.text()  # Store the filename
        self.has_unsaved_changes = False
        self.update_window_title()
        bg_filepath = os.path.join(self.input_dir, "inputs", self.current_filename)
        drawing_filepath = os.path.join(
            self.input_dir, "targets", self.current_filename
        )

        # Load background using helper method
        self.load_background_from_path(bg_filepath)

        # Load targets drawing if exists, otherwise create blank transparent layer
        if os.path.exists(drawing_filepath):
            self.load_image_from_path(drawing_filepath)
        else:
            self.image = QImage(self.background_image.size(), QImage.Format_ARGB32)
            self.image.fill(Qt.transparent)

        # Add initial state to undo manager without marking as unsaved
        self.undo_manager.push(self.image.copy())

        self.update_scene()

    # ---- Tool Operations ----

    def select_tool(self, tool):
        """Select and configure the current drawing tool"""
        # Save current tool settings before switching
        if self.current_tool:
            self.state.save_tool_state(
                self.current_tool,
                {
                    "size": self.graphics_view.pen_size,
                    "color": self.graphics_view.pen_color,
                },
            )

        self.graphics_view.clear_temp_line()
        self.current_tool = tool
        self.graphics_view.current_tool = tool

        # Restore saved settings for the selected tool
        tool_settings = self.state.restore_tool_state(tool)
        if tool_settings:
            self.graphics_view.pen_size = tool_settings["size"]
            self.graphics_view.pen_color = tool_settings["color"]

            # Update UI with tool settings
            self.ui.update_size_controls(tool_settings["size"])
            self.ui.update_color_button(tool_settings["color"])

        # Update button states
        buttons = {
            "Pencil": self.ui.pencil_button,
            "Eraser": self.ui.eraser_button,
            "Line": self.ui.line_button,
            "Fill": self.ui.fill_button,
        }

        # Set checked state for buttons
        for t, button in buttons.items():
            button.setChecked(t == tool)

        # Update cursor
        self.update_tool_cursor(tool)

    def update_tool_cursor(self, tool):
        """Update cursor based on selected tool"""
        if tool == "Eraser":
            self.graphics_view.update_eraser_cursor()
        elif tool == "Pencil":
            self.graphics_view.update_pencil_cursor()
        else:
            self.graphics_view.setCursor(self.graphics_view.tool_cursors[tool])

    # ---- Image Processing Operations ----

    def smooth_image(self):
        """Apply smoothing to the current image"""
        if self.image:
            self.save_undo_state()
            self.image = ImageProcessor.smooth_image(
                self.image, self.graphics_view.pen_color
            )
            self.update_scene()

    def thin_image(self):
        """Apply thinning to the current image"""
        if self.image:
            self.save_undo_state()
            self.image = ImageProcessor.thin_image(
                self.image, self.graphics_view.pen_color
            )
            self.update_scene()

    def smooth_thin_smooth_image(self):
        """Apply smooth-thin-smooth filter to the current image"""
        if self.image:
            self.save_undo_state()
            self.image = ImageProcessor.smooth_thin_smooth_image(
                self.image, self.graphics_view.pen_color
            )
            self.update_scene()

    # ---- Style Operations ----

    def choose_color(self):
        """Open color picker dialog"""
        if self.current_tool != "Eraser":  # Prevent changing eraser color
            color = QColorDialog.getColor()
            if color.isValid():
                self.state.update_tool_setting(self.current_tool, "color", color)
                self.ui.update_color_button(color)
                self.graphics_view.pen_color = color

    def update_color_button(self, color):
        """Deprecated: Use ui.update_color_button instead"""
        self.ui.update_color_button(color)

    def change_pen_size(self, size):
        """Update pen size and cursor"""
        # Store size in current tool's settings
        self.state.update_tool_setting(self.current_tool, "size", size)
        self.graphics_view.pen_size = size
        if self.current_tool in ["Eraser", "Pencil"]:
            getattr(self.graphics_view, f"update_{self.current_tool.lower()}_cursor")()

    def change_opacity(self, value):
        """Update the opacity of the drawing layer"""
        if self.pixmap_item:
            opacity = value / 100.0
            self.pixmap_item.setOpacity(opacity)

    # ---- Scene Management ----

    def update_scene(self):
        """Update the graphics scene with current layers"""
        if not self.scene:
            return

        self.scene.clear()

        # Add background layer
        if self.background_image:
            bg_pixmap = QPixmap.fromImage(self.background_image)
            self.background_item = self.scene.addPixmap(bg_pixmap)

        # Add drawing layer with opacity
        if self.image:
            pixmap = QPixmap.fromImage(self.image)
            self.pixmap_item = self.scene.addPixmap(pixmap)
            opacity = self.ui.opacity_slider.value() / 100.0
            self.pixmap_item.setOpacity(opacity)

        # Update view
        self.graphics_view.setSceneRect(self.scene.itemsBoundingRect())

        # Ensure zoom label is updated after scene changes
        self.graphics_view.update_zoom_label()

    # ---- State Management ----

    def undo(self):
        """Handle undo request"""
        if not self.undo_manager.can_undo():
            return

        prev_state = self.undo_manager.undo()
        if prev_state:
            self.image = prev_state
            self.update_scene()

    def redo(self):
        """Handle redo request"""
        state = self.undo_manager.redo()
        if state:
            self.image = state.copy()
            self.update_scene()

    def save_undo_state(self):
        """Save current state for undo"""
        if self.image:
            self.undo_manager.push(self.image.copy())
            self.has_unsaved_changes = True
            self.update_window_title()

    def update_window_title(self):
        """Update window title with filename and save status"""
        title = f"Drawing Tool - {self.state.current_filename}"
        if self.state.has_unsaved_changes:
            title += " *"
        self.setWindowTitle(title)

    def open_directory(self):
        """Open a directory dialog and reinitialize app with selected directory"""
        directory = QFileDialog.getExistingDirectory(
            self,
            "Open Directory",
            self.input_dir,
            QFileDialog.ShowDirsOnly | QFileDialog.DontResolveSymlinks,
        )

        if directory:
            # Save current settings
            current_settings = {
                "size": self.graphics_view.pen_size,
                "color": self.graphics_view.pen_color,
                "tool": self.current_tool,
                "opacity": self.ui.opacity_slider.value(),
            }

            # Update input directory
            self.input_dir = directory

            # Create necessary subdirectories
            os.makedirs(os.path.join(directory, "inputs"), exist_ok=True)
            os.makedirs(os.path.join(directory, "targets"), exist_ok=True)

            # Clear current image and scene
            self.image = None
            self.background_image = None
            self.current_filename = None
            self.has_unsaved_changes = False

            # Update scene
            self.scene.clear()

            # Reload images list
            self.load_images_list()

            # Restore settings
            self.graphics_view.pen_size = current_settings["size"]
            self.graphics_view.pen_color = current_settings["color"]
            self.select_tool(current_settings["tool"])
            self.ui.opacity_slider.setValue(current_settings["opacity"])

            # Update window title
            self.update_window_title()



class UndoManager:
    def __init__(self, max_steps=20):
        self.undo_stack = []
        self.redo_stack = []
        self.max_steps = max_steps

    def push(self, image):
        """Add new state to undo stack"""
        if image is None:
            return
        self.undo_stack.append(image.copy())
        # Clear redo stack when new action is performed
        self.redo_stack.clear()
        # Keep stack size under limit
        if len(self.undo_stack) > self.max_steps:
            self.undo_stack.pop(0)

    def undo(self):
        """Move current state to redo stack and return previous state"""
        if not self.undo_stack:
            return None
        state = self.undo_stack.pop()
        if state:
            self.redo_stack.append(state)
        return self.undo_stack[-1].copy() if self.undo_stack else None

    def redo(self):
        """Restore state from redo stack"""
        if not self.redo_stack:
            return None
        state = self.redo_stack.pop()
        if state:
            self.undo_stack.append(state)
        return state

    def can_undo(self):
        """Check if undo is available"""
        return len(self.undo_stack) > 1  # Need at least 2 states to undo

    def can_redo(self):
        """Check if redo is available"""
        return len(self.redo_stack) > 0



import numpy as np
from PyQt5.QtGui import QImage
from skimage.filters import gaussian
from skimage.morphology import thin
from .image_utils import ImageUtils


class ImageProcessor:
    @staticmethod
    def smooth_image(image, pen_color):
        """Apply Gaussian smoothing to the image"""
        if not image:
            return None

        # Convert to numpy array
        image_array = ImageUtils.qimage_to_numpy(image)

        # Create binary mask based on alpha channel (1 where alpha > 0)
        mask = (image_array[:, :, 3] > 0).astype(np.float32)

        # Apply Gaussian smoothing to the mask
        smoothed_mask = gaussian(mask, sigma=1)

        # Create new RGBA array
        height, width = mask.shape
        result = np.zeros((height, width, 4), dtype=np.uint8)

        # Where smoothed mask > 0, set the color and alpha
        mask_threshold = smoothed_mask > 0.1
        result[mask_threshold] = [
            pen_color.red(),
            pen_color.green(),
            pen_color.blue(),
            255,
        ]

        # Convert back to QImage and update
        return QImage(
            result.data, width, height, width * 4, QImage.Format_RGBA8888
        ).copy()

    @staticmethod
    def thin_image(image, pen_color):
        """Apply thinning to the image"""
        if not image:
            return None

        # Convert to numpy array
        image_array = ImageUtils.qimage_to_numpy(image)

        # Create binary mask based on alpha channel
        mask = (image_array[:, :, 3] > 0).astype(np.uint8)

        # Apply thinning
        thinned_mask = thin(mask)

        # Create new RGBA array
        height, width = mask.shape
        result = np.zeros((height, width, 4), dtype=np.uint8)

        # Where thinned mask is True, set the color and alpha
        result[thinned_mask] = [
            pen_color.red(),
            pen_color.green(),
            pen_color.blue(),
            255,
        ]

        # Convert back to QImage
        return QImage(
            result.data, width, height, width * 4, QImage.Format_RGBA8888
        ).copy()

    @staticmethod
    def create_binary_save_image(image):
        """Convert image to binary (black and white) for saving"""
        if not image:
            return None

        # Convert to numpy array
        image_array = ImageUtils.qimage_to_numpy(image)

        # Create a new image where alpha > 0 becomes white (255,255,255) and alpha = 0 becomes black (0,0,0)
        bw_array = np.zeros(
            (image_array.shape[0], image_array.shape[1], 3), dtype=np.uint8
        )
        bw_array[image_array[:, :, 3] > 0] = [255, 255, 255]

        # Convert back to QImage
        return QImage(
            bw_array.data,
            bw_array.shape[1],
            bw_array.shape[0],
            bw_array.shape[1] * 3,
            QImage.Format_RGB888,
        )

    @staticmethod
    def smooth_thin_smooth_image(image, pen_color):
        """Apply smooth-thin-smooth filter to the image"""
        if not image:
            return None

        # First smoothing
        for _ in range(2):
            image = ImageProcessor.smooth_image(image, pen_color)

        # Thinning
        image = ImageProcessor.thin_image(image, pen_color)

        # More smoothing passes
        for _ in range(4):
            image = ImageProcessor.smooth_image(image, pen_color)

        return image

    @staticmethod
    def flood_fill(image, start_point, fill_color):
        """Perform flood fill on the image starting from the given point"""
        if not image:
            return None

        # Make a copy of the original image
        result = image.copy()

        # Convert to numpy array just for the fill mask calculation
        image_array = ImageUtils.qimage_to_numpy(image)
        height, width = image_array.shape[:2]
        x, y = int(start_point.x()), int(start_point.y())

        # Bounds check
        if x < 0 or x >= width or y < 0 or y >= height:
            return image

        # Get target color (what we're replacing)
        target_alpha = image_array[y, x, 3]
        fill_alpha = fill_color.alpha()

        # If clicking on a pixel with the same transparency as fill color, return
        if (target_alpha > 0) == (fill_alpha > 0):
            return image

        # Create a mask for the fill area
        fill_mask = np.zeros((height, width), dtype=bool)

        # Stack for flood fill
        stack = [(x, y)]
        fill_mask[y, x] = True

        # Directions for 4-connected fill
        directions = [(0, 1), (1, 0), (0, -1), (-1, 0)]

        while stack:
            cx, cy = stack.pop()

            # Check neighboring pixels
            for dx, dy in directions:
                nx, ny = cx + dx, cy + dy

                # Check bounds and if pixel is unvisited
                if 0 <= nx < width and 0 <= ny < height and not fill_mask[ny, nx]:

                    # Check if the neighbor has the same alpha as target
                    if image_array[ny, nx, 3] == target_alpha:
                        stack.append((nx, ny))
                        fill_mask[ny, nx] = True

        # Only proceed if we're filling transparent areas
        if target_alpha == 0:
            # Paint directly on the QImage only where the mask is True
            for y in range(height):
                for x in range(width):
                    if fill_mask[y, x]:
                        result.setPixelColor(x, y, fill_color)

        return result



from PyQt5.QtGui import QColor, QImage
from PyQt5.QtCore import QObject, pyqtSignal


class AppState(QObject):
    # Signals for state changes
    tool_changed = pyqtSignal(str)
    pen_size_changed = pyqtSignal(int)
    pen_color_changed = pyqtSignal(QColor)
    opacity_changed = pyqtSignal(float)
    zoom_changed = pyqtSignal(float)
    image_changed = pyqtSignal()
    unsaved_changes = pyqtSignal(bool)

    def __init__(self):
        super().__init__()
        self._current_tool = "Line"
        self._tool_settings = {
            "Pencil": {"size": 8, "color": QColor(255, 165, 0)},  # Orange
            "Line": {"size": 8, "color": QColor(255, 165, 0)},  # Orange
            "Fill": {
                "size": 1,  # Size doesn't affect fill tool
                "color": QColor(255, 165, 0),  # Orange
            },
            "Eraser": {"size": 32, "color": QColor(0, 0, 0, 0)},  # Transparent
        }
        self._opacity = 0.5
        self._zoom_factor = 1.0
        self._current_image = None
        self._background_image = None
        self._has_unsaved_changes = False
        self._current_filename = "mask-out.png"

    # Tool property
    @property
    def current_tool(self):
        return self._current_tool

    @current_tool.setter
    def current_tool(self, tool):
        if self._current_tool != tool:
            self._current_tool = tool
            self.tool_changed.emit(tool)

    @property
    def current_tool_settings(self):
        return self._tool_settings[self._current_tool]

    def update_tool_setting(self, tool, setting, value):
        if tool in self._tool_settings and setting in self._tool_settings[tool]:
            self._tool_settings[tool][setting] = value
            if tool == self._current_tool:
                if setting == "size":
                    self.pen_size_changed.emit(value)
                elif setting == "color":
                    self.pen_color_changed.emit(value)

    def save_tool_state(self, tool, settings):
        """Save current settings for the given tool"""
        if tool in self._tool_settings:
            self._tool_settings[tool] = settings.copy()

    def restore_tool_state(self, tool):
        """Restore saved settings for the given tool"""
        if tool in self._tool_settings:
            return self._tool_settings[tool].copy()
        return None

    # Pen size property
    @property
    def pen_size(self):
        return self._pen_size

    @pen_size.setter
    def pen_size(self, size):
        if self._pen_size != size:
            self._pen_size = size
            self.pen_size_changed.emit(size)

    # Pen color property
    @property
    def pen_color(self):
        return self._pen_color

    @pen_color.setter
    def pen_color(self, color):
        if self._pen_color != color:
            self._pen_color = color
            self.pen_color_changed.emit(color)

    # Opacity property
    @property
    def opacity(self):
        return self._opacity

    @opacity.setter
    def opacity(self, value):
        if self._opacity != value:
            self._opacity = value
            self.opacity_changed.emit(value)

    # Zoom factor property
    @property
    def zoom_factor(self):
        return self._zoom_factor

    @zoom_factor.setter
    def zoom_factor(self, factor):
        if self._zoom_factor != factor:
            self._zoom_factor = factor
            self.zoom_changed.emit(factor)

    # Current image property
    @property
    def current_image(self):
        return self._current_image

    @current_image.setter
    def current_image(self, image):
        self._current_image = image
        self.image_changed.emit()

    # Background image property
    @property
    def background_image(self):
        return self._background_image

    @background_image.setter
    def background_image(self, image):
        self._background_image = image
        self.image_changed.emit()

    # Current filename property
    @property
    def current_filename(self):
        return self._current_filename

    @current_filename.setter
    def current_filename(self, filename):
        self._current_filename = filename

    # Unsaved changes property
    @property
    def has_unsaved_changes(self):
        return self._has_unsaved_changes

    @has_unsaved_changes.setter
    def has_unsaved_changes(self, value):
        if self._has_unsaved_changes != value:
            self._has_unsaved_changes = value
            self.unsaved_changes.emit(value)

    def create_blank_image(self, size):
        """Create a new blank transparent image"""
        image = QImage(size, QImage.Format_ARGB32)
        image.fill(QColor(0, 0, 0, 0))
        return image



import numpy as np
from PyQt5.QtGui import QImage


class ImageUtils:
    @staticmethod
    def qimage_to_numpy(qimage):
        """Convert QImage to numpy array safely accounting for bytes per line."""
        width = qimage.width()
        height = qimage.height()
        bytes_per_line = qimage.bytesPerLine()

        ptr = qimage.bits()
        ptr.setsize(height * bytes_per_line)
        arr = np.frombuffer(ptr, np.uint8)

        # Reshape considering 4 channels (RGBA)
        if bytes_per_line == width * 4:
            return arr.reshape((height, width, 4))

        # If there's padding, we need to remove it
        return arr.reshape(height, bytes_per_line // 4, 4)[:, :width]

    @staticmethod
    def numpy_to_qimage(array):
        """Convert numpy array to QImage ensuring proper alignment."""
        height, width, channels = array.shape
        array = np.require(array, np.uint8, "C")  # Ensure contiguous memory
        qimage = QImage(array.data, width, height, width * 4, QImage.Format_RGB32)
        return qimage.copy()  # Create a deep copy to ensure data ownership



from PyQt5.QtWidgets import QGraphicsView
from PyQt5.QtGui import QPainter, QPen, QPixmap, QCursor, QTransform
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QPainter, QPen, QColor
from .image_processor import ImageProcessor  # Add this import


class CustomGraphicsView(QGraphicsView):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.drawing = False
        self.panning = False
        self.last_point = None
        self.last_pan_pos = None
        self.current_tool = "Line"
        self.main_window = parent
        self.temp_line = None
        self.setDragMode(QGraphicsView.NoDrag)
        # Enable scrollbars
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.pen_size = 8
        self.pen_color = QColor(255, 165, 0)  # Orange color (RGB: 255, 165, 0)
        # Add cursor definitions
        self.tool_cursors = {
            "Pencil": Qt.CustomCursor,
            "Line": Qt.CrossCursor,
            "Eraser": Qt.CustomCursor,
            "Fill": Qt.PointingHandCursor,
        }
        self.update_tool_cursor()
        self.setFocusPolicy(Qt.StrongFocus)
        self.zoom_factor = 1.0
        self.min_zoom = 0.1
        self.max_zoom = 5.0
        # Enable mouse tracking for wheel events
        self.setMouseTracking(True)

    def clear_temp_line(self):
        if self.temp_line:
            if self.temp_line.scene():
                self.temp_line.scene().removeItem(self.temp_line)
            self.temp_line = None

    def update_pencil_cursor(self):
        """Update the pencil cursor size"""
        if self.current_tool == "Pencil":
            size = self.pen_size + 2
            pixmap = QPixmap(size + 2, size + 2)
            pixmap.fill(Qt.transparent)
            painter = QPainter(pixmap)

            # Draw circle
            painter.setPen(QPen(Qt.black, 1))
            painter.drawEllipse(1, 1, size - 1, size - 1)

            # Draw cross
            center = size // 2 + 1
            cross_size = min(size // 3, 4)  # Limit cross size
            painter.drawLine(center - cross_size, center, center + cross_size, center)
            painter.drawLine(center, center - cross_size, center, center + cross_size)

            painter.end()
            self.setCursor(QCursor(pixmap))

    def update_eraser_cursor(self):
        """Update the eraser cursor size"""
        if self.current_tool == "Eraser":
            size = self.pen_size
            pixmap = QPixmap(size + 2, size + 2)
            pixmap.fill(Qt.transparent)
            painter = QPainter(pixmap)
            painter.setPen(QPen(Qt.black, 1))
            painter.drawEllipse(1, 1, size - 1, size - 1)
            painter.end()
            self.setCursor(QCursor(pixmap))

    def update_tool_cursor(self):
        """Update cursor based on current tool"""
        if self.current_tool == "Pencil":
            self.update_pencil_cursor()
        elif self.current_tool == "Eraser":
            self.update_eraser_cursor()
        elif self.current_tool == "Line":
            self.setCursor(self.tool_cursors["Line"])

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton and self.scene() and self.main_window.image:
            # Clear any existing temporary line
            self.clear_temp_line()
            self.drawing = True
            self.last_point = self.mapToScene(event.pos())

            if self.current_tool == "Fill":
                # Handle fill tool
                try:
                    new_image = ImageProcessor.flood_fill(
                        self.main_window.image, self.last_point, self.pen_color
                    )
                    if new_image:
                        self.main_window.image = new_image
                        self.main_window.update_scene()
                        self.main_window.save_undo_state()
                except Exception as e:
                    print(f"Fill error: {e}")
            elif self.current_tool == "Line":
                # Create temporary line item for preview
                self.temp_line = self.scene().addLine(
                    self.last_point.x(),
                    self.last_point.y(),
                    self.last_point.x(),
                    self.last_point.y(),
                    QPen(self.pen_color, self.pen_size, Qt.SolidLine),
                )
        elif event.button() == Qt.MiddleButton:
            self.panning = True
            self.last_pan_pos = event.pos()
            self.setCursor(Qt.ClosedHandCursor)
            event.accept()

    def mouseMoveEvent(self, event):
        if self.drawing and self.scene() and self.main_window.image:
            current_point = self.mapToScene(event.pos())
            if self.last_point and not current_point.isNull():
                if self.current_tool == "Line":
                    # Update temporary line preview
                    if self.temp_line:
                        pen = QPen(self.pen_color, self.pen_size, Qt.SolidLine)
                        pen.setCapStyle(Qt.RoundCap)
                        pen.setJoinStyle(Qt.RoundJoin)
                        self.temp_line.setLine(
                            self.last_point.x(),
                            self.last_point.y(),
                            current_point.x(),
                            current_point.y(),
                        )
                        self.temp_line.setPen(pen)
                else:
                    try:
                        painter = QPainter(self.main_window.image)
                        if self.current_tool == "Eraser":
                            # Set composition mode to clear for eraser
                            painter.setCompositionMode(QPainter.CompositionMode_Clear)
                            pen = QPen(Qt.transparent, self.pen_size, Qt.SolidLine)
                        else:
                            painter.setCompositionMode(
                                QPainter.CompositionMode_SourceOver
                            )
                            pen = QPen(self.pen_color, self.pen_size, Qt.SolidLine)

                        # Set round cap and join for smoother lines
                        pen.setCapStyle(Qt.RoundCap)
                        pen.setJoinStyle(Qt.RoundJoin)
                        painter.setPen(pen)

                        painter.drawLine(
                            self.last_point.toPoint(), current_point.toPoint()
                        )
                        painter.end()
                        self.last_point = current_point
                        self.main_window.update_scene()
                    except Exception as e:
                        print(f"Drawing error: {e}")
                        self.drawing = False
                        self.last_point = None
        elif self.panning and self.last_pan_pos is not None:
            delta = event.pos() - self.last_pan_pos
            self.horizontalScrollBar().setValue(
                self.horizontalScrollBar().value() - delta.x()
            )
            self.verticalScrollBar().setValue(
                self.verticalScrollBar().value() - delta.y()
            )
            self.last_pan_pos = event.pos()
            self.last_point = None
        elif self.panning and self.last_pan_pos is not None:
            delta = event.pos() - self.last_pan_pos
            self.horizontalScrollBar().setValue(
                self.horizontalScrollBar().value() - delta.x()
            )
            self.verticalScrollBar().setValue(
                self.verticalScrollBar().value() - delta.y()
            )
            self.last_pan_pos = event.pos()
            event.accept()

    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            if self.drawing:
                if self.current_tool == "Line" and self.last_point:
                    # Draw the final line on the image
                    try:
                        # Clear temp line first
                        self.clear_temp_line()

                        current_point = self.mapToScene(event.pos())
                        # Draw the final line
                        painter = QPainter(self.main_window.image)
                        pen = QPen(self.pen_color, self.pen_size, Qt.SolidLine)
                        pen.setCapStyle(Qt.RoundCap)
                        pen.setJoinStyle(Qt.RoundJoin)
                        painter.setPen(pen)
                        painter.drawLine(
                            self.last_point.toPoint(), current_point.toPoint()
                        )
                        painter.end()

                        # Update the scene after drawing
                        self.main_window.update_scene()
                    except Exception as e:
                        print(f"Drawing error: {e}")

                # Save state after any drawing operation completes
                self.main_window.save_undo_state()

            # Reset drawing state
            self.drawing = False
            self.last_point = None

        elif event.button() == Qt.MiddleButton:
            self.panning = False
            self.last_pan_pos = None
            self.setCursor(Qt.ArrowCursor)
            event.accept()

    def keyPressEvent(self, event):
        # Handle Ctrl+Z for undo
        if event.modifiers() == Qt.ControlModifier:
            if event.key() == Qt.Key_Z:
                self.main_window.undo()
            elif event.key() == Qt.Key_S:
                self.main_window.save_image()
            elif event.key() == Qt.Key_0:  # Add Ctrl+0 shortcut
                self.zoom_100()
        super().keyPressEvent(event)

    def scale_view(self, factor):
        """Scale the view by the given factor"""
        new_zoom = self.zoom_factor * factor
        if self.min_zoom <= new_zoom <= self.max_zoom:
            self.zoom_factor = new_zoom
            self.setTransform(QTransform().scale(self.zoom_factor, self.zoom_factor))
            self.update_zoom_label()

    def zoom_in(self):
        """Zoom in by 20%"""
        self.scale_view(1.2)

    def zoom_out(self):
        """Zoom out by 20%"""
        self.scale_view(1/1.2)

    def zoom_fit(self):
        """Fit view to scene contents"""
        if self.scene() and not self.scene().items():
            return
        self.fitInView(self.scene().itemsBoundingRect(), Qt.KeepAspectRatio)
        # Update zoom factor based on current transform
        self.zoom_factor = self.transform().m11()
        self.update_zoom_label()

    def zoom_100(self):
        """Reset zoom to 100%"""
        self.zoom_factor = 1.0
        self.setTransform(QTransform().scale(1.0, 1.0))
        self.update_zoom_label()

    def update_zoom_label(self):
        """Update the zoom level display in the toolbar"""
        if hasattr(self.main_window, "ui"):
            zoom_percent = int(self.zoom_factor * 100)
            self.main_window.ui.zoom_level_label.setText(f"{zoom_percent}%")

    def wheelEvent(self, event):
        """Handle mouse wheel for zooming"""
        if event.modifiers() == Qt.ControlModifier:
            # Get the position before scaling to use as anchor point
            anchor_pos = self.mapToScene(event.pos())

            # Calculate zoom factor based on wheel delta
            delta = event.angleDelta().y()
            factor = 1.1 if delta > 0 else 1 / 1.1

            # Check if the new zoom would be within bounds
            new_zoom = self.zoom_factor * factor
            if self.min_zoom <= new_zoom <= self.max_zoom:
                self.zoom_factor = new_zoom
                # Apply the transform
                transform = QTransform()
                transform.scale(self.zoom_factor, self.zoom_factor)
                self.setTransform(transform)

                # Adjust the view to maintain the point under cursor
                new_pos = self.mapFromScene(anchor_pos)
                delta = new_pos - event.pos()
                self.horizontalScrollBar().setValue(
                    self.horizontalScrollBar().value() + delta.x()
                )
                self.verticalScrollBar().setValue(
                    self.verticalScrollBar().value() + delta.y()
                )

                # Update zoom display
                self.update_zoom_label()

            event.accept()
        else:
            # Handle normal scrolling
            super().wheelEvent(event)



from setuptools import setup, find_packages

setup(
    name="ortho-editor",
    version="0.1.13",
    packages=find_packages(),
    include_package_data=True,
    package_data={
        'ortho_editor': ['fonts/*.ttf'],
    },
    install_requires=[
        "black",
        "flake8",
        "PyQt5",
    ],
    entry_points={
        "console_scripts": [
            "ortho-editor=ortho_editor.editor:main",
        ],
    },
    author="yaacov",
    author_email="kobi.zamir@gmail.com",
    description="ORTHO mask editor tool",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/yaacov/ortho-editor",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)
